/* N-body calculation using Greenspan 1st-order discrete mechanics. */
/* Incorporates xyz-parallelism, uses domain pointers, combines     */
/* particle and pair domains, diagonal processors store particles   */
/* ***  Includes communication operations only - arithmetic omitted */

#include <stdio.hs>
/* #include <math.hs> */
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs. */
#define N 64

int NB;      /* Number of bodies in system, <= N */
int LNB;     /* Largest power of 2 < NB */

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/
/* act  - processor is active pair					*/
/* sm   - processor participates in acceleration sum			*/
/* here - processor represents particle					*/

domain pair { int i, j, act, sm, here;
	      float x0, x1, v0, v1, m, d0, r0, ac;
	      domain pair *pi, *pj;  domain pair *r; };

domain pair P[3][N][N];

/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.   */

void pair::setup()
 { int offset, c;
   for ( LNB = 1; LNB < NB; LNB <<= 1 ) ;  LNB >>= 1;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);
   act = sm && (i != j);  here = sm && (i == j);
   pi = &P[c][i][i];  pj = &P[c][j][j];  r = (c<2) ? this+N*N : this-2*N*N;
 }

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { [domain pair].
    if (here) 
     { x0 = 17.5;  x1 = -23.37;  v0 = 5.14;  m = 25.6;  d0 = 14.0;
       r0 = 3.6;  ac = 4.76; }
 }

/* Calculate accelerations */

void pair::accel()
 {    float d1, r1;  int smk, smm;
      mono int k;
      if (act)
      { d1 = pj->x1;  d1 = pi->x1;
	r1 = d1;  ac = r->r1;  r1 = r->ac; };
      /* Sum acceleration terms in time ~ O(log N) */
      smk = j;  smm = 1;
      if (sm)
       for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (! (smk & 1)) && ( j+k < NB) )
	     { ac = (this+k)->ac;  smk >>= 1; }
	else smm = 0; }   /* end for, end if (sm) ... */
      if (j == 0) pi->ac = ac;
 }

/* Calculate initial accelerations */

void pair::accel1()
 {    int smk, smm;
      mono int k;
      if (act)
      { d0 = pj->x0;  d0 = pi->x0;
	r0 = d0;  ac = r->r0;  r0 = r->ac; };  /* end if (act) */
      /* Sum acceleration terms in time ~ O(log N) */
      smk = j;  smm = 1;
      if (sm)
       for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (! (smk & 1)) && ( j+k < NB) )
	     { ac = (this+k)->ac;  smk >>= 1; }
	else smm = 0; }   /* end for, end if (sm) ... */
      if (j == 0) pi->ac = ac;
 }


/* Greenspan's 1st-order method.  Performs ns steps of time dt
   with solution tolerance eps.  Returns number of evaluations.
   Leaves number of iterations which were stopped at limit in cfail.
   Some printing for debugging is provided in comments.  */

int cfail;   /* Counts number of iterations that failed to converge */

int green1(ns, lim, dt, eps)
  int ns, lim;  float dt, eps;
 { int k, ic, v;  float hdt, mx, my;
   v = 0;  hdt = 0.005;
   [domain pair].
   { float nx;
     if (act) m = pj->m;
     for (k = 0; k < ns; k++)
       { if (here) { x1 = x0;  v1 = v0;  ic = 0; };
         accel1();
	 /* This is a do - while - repeat loop: */
   	 do { v++;  ic++;
	      if ( ! ( ic < lim ) ) break;
	      accel(); }
	 while (1);
       } ;   /* end for */
    }  /* end  [domain pair] */
   return(v);
 }

/* Commands are:
   r - run (prompts for parameters)
   q - quit                         */

void main()
 { int ns, np, nl, nv, k, vl;  float dt, eps;  long tm;
   char c[2];
   while ( printf("nbcmtim> "), scanf("%1s", c), c[0] != 'q' )
    {
     /* Performs np integrations, each ns steps of dt */
     printf(
         "Enter NB, no. of evals per step, no. of steps, no. of cycles.\n> ");
     scanf("%d%d%d%d", &NB, &nv, &ns, &np);
     [domain pair].{ setup(); };
     init();
     printf("\n time = 0");
     for ( k = 0; k < np; k++ )
      { CM_start_timer(1);
        cfail = 0;
        vl = green1(ns, nv, dt, eps);
        CM_stop_timer(1);  CM_reset_timer();
        printf(
	 "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	  (k + 1)*ns*dt, vl, cfail);
      };
    }
 }

